# author:    Adrian Rosebrock
# website:   http://www.pyimagesearch.com

# import the necessary packages
from scipy.spatial import distance as dist
import numpy as np
import cv2

def order_points(pts):
    # sort the points based on their x-coordinates
    xSorted = pts[np.argsort(pts[:, 0]), :]

    # grab the left-most and right-most points from the sorted
    # x-roodinate points
    leftMost = xSorted[:2, :]
    rightMost = xSorted[2:, :]

    # now, sort the left-most coordinates according to their
    # y-coordinates so we can grab the top-left and bottom-left
    # points, respectively
    leftMost = leftMost[np.argsort(leftMost[:, 1]), :]
    (tl, bl) = leftMost

    # now that we have the top-left coordinate, use it as an
    # anchor to calculate the Euclidean distance between the
    # top-left and right-most points; by the Pythagorean
    # theorem, the point with the largest distance will be
    # our bottom-right point
    D = dist.cdist(tl[np.newaxis], rightMost, "euclidean")[0]
    (br, tr) = rightMost[np.argsort(D)[::-1], :]

    # return the coordinates in top-left, top-right,
    # bottom-right, and bottom-left order
    return np.array([tl, tr, br, bl], dtype="float32")

def four_point_transform(image, pts, bvalue): # Edit: add param bvalue so we can use custom border color (instead of default black)
    # obtain a consistent order of the points and unpack them
    # individually
    rect = order_points(pts)
    (tl, tr, br, bl) = rect

    # compute the width of the new image, which will be the
    # maximum distance between bottom-right and bottom-left
    # x-coordiates or the top-right and top-left x-coordinates
    widthA = np.sqrt(((br[0] - bl[0]) ** 2) + ((br[1] - bl[1]) ** 2))
    widthB = np.sqrt(((tr[0] - tl[0]) ** 2) + ((tr[1] - tl[1]) ** 2))
    maxWidth = max(int(widthA), int(widthB))

    # compute the height of the new image, which will be the
    # maximum distance between the top-right and bottom-right
    # y-coordinates or the top-left and bottom-left y-coordinates
    heightA = np.sqrt(((tr[0] - br[0]) ** 2) + ((tr[1] - br[1]) ** 2))
    heightB = np.sqrt(((tl[0] - bl[0]) ** 2) + ((tl[1] - bl[1]) ** 2))
    maxHeight = max(int(heightA), int(heightB))

    # now that we have the dimensions of the new image, construct
    # the set of destination points to obtain a "birds eye view",
    # (i.e. top-down view) of the image, again specifying points
    # in the top-left, top-right, bottom-right, and bottom-left
    # order
    dst = np.array([
        [0, 0],
        [maxWidth - 1, 0],
        [maxWidth - 1, maxHeight - 1],
        [0, maxHeight - 1]], dtype="float32")

    # compute the perspective transform matrix and then apply it
    M = cv2.getPerspectiveTransform(rect, dst)
    warped = cv2.warpPerspective(image, M, (maxWidth, maxHeight), borderValue=bvalue)

    # return the warped image
    return warped


#!/usr/bin/env python3

config_raw = r'''# NOTE: on windows backslash needs to be escaped like so:
# input: "C:\\my path\\to\\input"
# output: "C:\\my path\\to\\output"

# Project name
# This is used to create an unique path for your project inside output directory
# Each slice run creates a new unique path (project_name+timestamp)
# Example: output/MyProject_2024_06_22_17_48_45
project-name: "MyProject"

# Path to input directory
# Create subfolders inside this directory to organize your scans
# Each subfolder adds a layer to the naming scheme
# Example: input/Birthdays/2010/Michael -> Birthdays_2010_Michael_1.jpg, Birthdays_2010_Michael_2.jpg ...
# Each subfolder is also mirrored in the output directory for convenience
input: "/path/to/input/"

# Path to output directory
# Your slices will be placed here under unique directory
output: "/path/to/output/"

# Number of workers used for multiprocessing
# Use half of physical cpu cores as a safe default value
workers: 2

# Skip the need to confirm action modes (True/False)
skip-confirm: False

# Detection sensitivity (1-255)
# Try to match this with the color between scanned items (in grayscale)
white-threshold: 230

# Minimum size of the sliced image in % compared to scanned image (1-100)
minimum-size: 1.00

# Maximum size of the sliced image in % compared to scanned image (1-100)
maximum-size: 45.00

# Enable count mode (True/False)
count-mode: False

# Enable test mode (True/False)
test-mode: False

# Enable preview mode (True/False)
preview-mode: False

# Enable slice mode (True/False)
slice-mode: False

# Scale using factor value (e.g: 0.5, 1.5, 50.0)
# 0 = disabled
scale-factor: 0

# Scale using new width value in pixels
# Image will not be upscaled
# Keeps aspect ratio
# 0 = disabled
scale-width: 0

# Scale using new height value in pixels
# Image will not be upscaled
# Keeps aspect ratio
# 0 = disabled
scale-height: 0

# Remove noise from slice (0-5)
# Anything over 3 is slow
# 0 = disabled
filter-denoise: 1

# Add color (values over 1.0)
# Remove color (values under 1.0)
filter-color: 1.0

# Add contrast (values over 1.0)
# Remove contrast (values under 1.0)
filter-contrast: 1.0

# Add brightness (values over 1.0)
# Remove brightness (values under 1.0)
filter-brightness: 1.0

# Add sharpness (values over 1.0)
# Remove sharpness (values under 1.0)
filter-sharpness: 1.0

# Adjust LUT strength value (0.0 - 1.0)
# 0.0 returns the original image while 1.0 returns the LUT image
filter-lut-strength: 1.0

# Path to the LUT file
# Has to be in cube format (.cube)
# "" = disabled
filter-lut-path: ""

# Use perspective fix (0-89)
# Maximum allowed tilt value of the sliced image in degrees
# 0 = disabled
perspective-fix: 0

# Use auto-rotate (disable/cw/ccw)
# Rotates slices automatically 90 degrees CW or CCW
# if the slice width is smaller than it's height
auto-rotate: "disable"

# Save format (png|jpeg|webp)
save-format: "jpeg"

# Optimize PNG files (True/False)
png-optimize: True

# Level of PNG compression (0-9)
# 0 = disabled
png-compression: 3

# Optimize JPEG files (True/False)
jpeg-optimize: True

# Level of JPEG quality (0-100)
jpeg-quality: 95

# Save WEBP as lossless image (True/False)
webp-lossless: False

# WEBP saving file quality/speed tradeoff (0-6)
# 0=fast, 6=slower-better
webp-method: 4

# Level of WEBP quality (0-100)
# Note: if you enable lossless webp then this value becomes the compression rate
webp-quality: 90

# GUI font scale
font-scale: 1.0

# GUI theme
# Check FreeSimpleGUI github page for theme names, some examples:
# Light: LightGray1, Material2, SystemDefault
# Dark: Black, Dark, DarkGrey5
theme: "DarkBlack"

# Max image width inside GUI
view-width: 960

# Max image height inside GUI
view-height: 800

'''


#!/usr/bin/env python3

import os
import sys
import multiprocessing

from .confparser import conf_parser_p
from .scis import *
from .utils import create_statusline
from .scis_logger import *

class Param:
    pass

def main():
    p = Param()
    tasks = []
    queue = multiprocessing.Manager().Queue(-1)

    # Create path for config file based on platform
    if sys.platform in ["win32", "cygwin"]:
        p.path_config_dir = os.path.join(os.path.expanduser("~"), "Documents", "ScanImageSlicer")
    elif sys.platform == "darwin":
        p.path_config_dir = os.path.join(os.path.expanduser("~"), "Library", "Application Support", "ScanImageSlicer")
    elif sys.platform == "linux":
        p.path_config_dir = os.path.join(os.path.expanduser("~"), ".config", "ScanImageSlicer")
    else:
        p.path_config_dir = os.path.join(os.path.expanduser("~"), ".ScanImageSlicer")

    p.log_path = os.path.join(p.path_config_dir, "last_run.log")

    # Create the default config file directory if needed
    if not os.path.exists(p.path_config_dir):
        os.mkdir(p.path_config_dir)

    listener = multiprocessing.Process(
        target=listener_process,
        args=(p, queue, listener_configurer)
    )

    listener.start()
    logger = queue_configurer(queue)

    # Collect the program params
    p = conf_parser_p(p)

    if p.cont:
        # Print the statusline
        print(create_statusline(p.name, p.version, p.description, p.path_config_file), flush=True)

        # Parse params
        p = parse_p(p)

    if p.cont:
        # Collect images from input path
        images = collect_images(p.input)

        if images:
            # List all compatible images from input directory
            if p.list_images:
                list_images(images)

            # Save a list of compatible images into output directory
            if p.list_file:
                save_imagelist_as_txt(p.output, p.project_name, images)

            # Add or remove tasks
            if any([p.add_all, p.add_id, p.add_new, p.add_old, p.add_random, p.remove_id]):
                tasks = handle_tasks(p, images)
                tasks = sorted(tasks, key=int)

            # List tasks
            if p.list_tasks:
                list_tasks(tasks, images)

            # Run tasks
            if any([p.test_mode, p.preview_mode, p.slice_mode, p.count_mode]):
                if tasks:
                    if confirm(p.skip_confirm, len(tasks), p.run_mode):
                        run_tasks(queue, p, tasks, images)

                        if p.slice_mode:
                            sequential_parallel_rename(p)
                else:
                    logger.info("Add some tasks before using action modes\n")
        else:
            logger.info(f"No compatible images found at: {p.input}\n")

    if p.run_mode:
        logger.info(f"View runlog at: {p.log_path}\n")

    # Stop queue logger
    queue.put_nowait(None)
    listener.join()

if __name__ == "__main__":
    sys.exit(main())


#!/usr/bin/env python3

import FreeSimpleGUI as sg

class Header(sg.Text):
    def __init__(self, p, text, **kw):
        scaled_font_size = int(p.font_scale * 11.0)
        kw.setdefault("p", ((2, 2), (10, 0)))
        kw.setdefault("font", ('Arial', scaled_font_size, "bold"))
        super().__init__(text=text, **kw)

class Text(sg.Text):
    def __init__(self, p, text, **kw):
        scaled_font_size = int(p.font_scale * 10.0)
        kw.setdefault("p", (2, 2))
        kw.setdefault("font", ('Arial', scaled_font_size))
        super().__init__(text=text, **kw)

class Spin(sg.Spin):
    def __init__(self, p, values, initial_value, **kw):
        scaled_font_size = int(p.font_scale * 10.0)
        kw.setdefault("p", (2, 2))
        kw.setdefault("size", (10, 1))
        kw.setdefault("font", ('Arial', scaled_font_size))
        kw.setdefault("enable_events", True)
        super().__init__(values=values, initial_value=initial_value, **kw)

class Input(sg.Input):
    def __init__(self, p, default_value, **kw):
        scaled_font_size = int(p.font_scale * 10.0)
        kw.setdefault("p", (2, 2))
        kw.setdefault("font", ('Arial', scaled_font_size))
        kw.setdefault("enable_events", True)
        super().__init__(default_text=default_value, **kw)

class Slider(sg.Slider):
    def __init__(self, p, type, range, default_value, **kw):
        scaled_font_size = int(p.font_scale * 10.0)

        if type == "int":
            kw.setdefault("range", range)
        elif type == "float":
            kw.setdefault("range", range)
            kw.setdefault("resolution", 0.1)

        kw.setdefault("p", (2, 0))
        kw.setdefault("orientation", "h")
        kw.setdefault("expand_x", True)
        kw.setdefault("enable_events", True)
        kw.setdefault("font", ('Arial', scaled_font_size))
        super().__init__(default_value=default_value, **kw)

class Button(sg.Button):
    def __init__(self, p, text, **kw):
        scaled_font_size = int(p.font_scale * 10.0)
        kw.setdefault("s", (19, 2))
        kw.setdefault("p", (2, 2))
        kw.setdefault("expand_x", True)
        kw.setdefault("font", ('Arial', scaled_font_size))
        super().__init__(button_text=text, **kw)

class Image(sg.Image):
    def __init__(self, source, **kw):
        super().__init__(source=source, **kw)

class Column(sg.Column):
    def __init__(self, layout, **kw):
        kw.setdefault("vertical_alignment", "top")
        super().__init__(layout=layout, **kw)

class Window(sg.Window):
    def __init__(self, title, layout, **kw):
        kw.setdefault("resizable", True)
        super().__init__(title=title, layout=layout, **kw)


#!/usr/bin/env python3

import os
import logging
import configargparse
from .__init__ import __doc__
from .__version__ import __version__
from .config import config_raw
from .utils import create_statusline

def conf_parser_p(p):
    logger = logging.getLogger()
    p.cont = True
    p.version = __version__
    p.name = "Scan-Image-Slicer"
    p.description = __doc__

    # Create the default config file if it does not exist
    p.name_config_file = "v" + __version__.replace(".", "") + "_config.yaml"
    p.path_config_file = os.path.join(p.path_config_dir, p.name_config_file)

    if not os.path.isfile(p.path_config_file):
        with open(p.path_config_file, 'w') as outfile:
            try:
                outfile.writelines(config_raw)
                outfile.close()
            except OSError as e:
                logger.error(e)
                p.cont = False

        # Notify the user about default config file being created
        if os.path.isfile(p.path_config_file):
            logger.info("Created default config file:")
            logger.info(p.path_config_file + "\n")
        else:
            # Exit if unable to create config file
            logger.error(f"Could not create default config file: {p.path_config_file}")
            p.cont = False

    if not p.cont:
        return p

    # Setup confargparser
    description = create_statusline(p.name, p.version, p.description, p.path_config_file)

    parser = configargparse.ArgumentParser(
        default_config_files=[p.path_config_file],
        description=description,
        config_file_parser_class=configargparse.YAMLConfigFileParser,
        formatter_class=configargparse.RawTextHelpFormatter,
        exit_on_error=False,
        allow_abbrev=True
    )

    parser.add_argument("-conf", "--config-file", metavar="FILE", is_config_file=True, help="Path to custom config file")
    parser.add_argument("-skip", "--skip-confirm", action="store_true", help="Skip the need to confirm action modes")
    parser.add_argument("-work", "--workers", metavar="NUM", type=int, help="Number of workers for multiprocessing")
    parser.add_argument("-name", "--project-name", metavar="TEXT", type=str, help="Project name")

    mode_group = parser.add_argument_group("Modes")
    mode_select_group = mode_group.add_mutually_exclusive_group()
    mode_select_group.add_argument("-test", "--test-mode", action="store_true", help="Enable test mode")
    mode_select_group.add_argument("-count", "--count-mode", action="store_true", help="Enable count mode")
    mode_select_group.add_argument("-preview", "--preview-mode", action="store_true", help="Enable preview mode")
    mode_select_group.add_argument("-slice", "--slice-mode", action="store_true", help="Enable slice mode")

    path_group = parser.add_argument_group("Paths")
    path_group.add_argument("-i", "--input", metavar="PATH", type=str, help="PATH to input directory")
    path_group.add_argument("-o", "--output", metavar="PATH", type=str, help="PATH to output directory")
    path_group.add_argument("-lutP", "--filter-lut-path", metavar="FILE", type=str, help="Path to lut .cube file")

    detect_group = parser.add_argument_group("Image slice detection")
    detect_group.add_argument("-white", "--white-threshold", metavar="NUM", type=int, help="White level between slices (1-255)")
    detect_group.add_argument("-min", "--minimum-size", metavar="NUM", type=float, help="Minimum slice size in %% (1-100)")
    detect_group.add_argument("-max", "--maximum-size", metavar="NUM", type=float, help="Maximum slice size in %% (1-100)")

    scale_group = parser.add_argument_group("Image slice scaling")
    scale_group.add_argument("-scaleF", "--scale-factor", metavar="NUM", type=float, help="Scale slice with factor value")
    scale_group.add_argument("-scaleW", "--scale-width", metavar="NUM", type=int, help="Scale slice to new width (no upscale)")
    scale_group.add_argument("-scaleH", "--scale-height", metavar="NUM", type=int, help="Scale slice to new height (no upscale)")

    filter_group = parser.add_argument_group("Image slice filters")
    filter_group.add_argument("-denoise", "--filter-denoise", metavar="NUM", type=int, help="Remove noise from slice (0-5)")
    filter_group.add_argument("-lutS", "--filter-lut-strength", metavar="NUM", type=float, help="Adjust LUT strength value (0.0-1.0)")
    filter_group.add_argument("-color", "--filter-color", metavar="NUM", type=float, help="Add or remove color from slice (0.0-2.0)")
    filter_group.add_argument("-contrast", "--filter-contrast", metavar="NUM", type=float, help="Add or remove contrast from slice (0.0-2.0)")
    filter_group.add_argument("-brightness", "--filter-brightness", metavar="NUM", type=float, help="Add or remove brightness from slice (0.0-2.0)")
    filter_group.add_argument("-sharpness", "--filter-sharpness", metavar="NUM", type=float, help="Add or remove sharpness from slice (0.0-2.0)")

    fixes_group = parser.add_argument_group("Image slice tweaks")
    fixes_group.add_argument("-pfix", "--perspective-fix", metavar="NUM", type=int, help="Fix slice tilt (0-89)")
    fixes_group.add_argument("-autoR", "--auto-rotate", metavar="TEXT", type=str, help="Rotate slice 90 deg CW/CCW if w < h")

    format_group = parser.add_argument_group("File format")
    format_group.add_argument("-save", "--save-format", metavar="TEXT", type=str, help="File format for slices")
    format_group.add_argument("-pngO", "--png-optimize", action="store_true", help="Try to optimize PNG file")
    format_group.add_argument("-pngC", "--png-compression", metavar="NUM", type=int, help="Try to compress PNG file (0-9)")
    format_group.add_argument("-jpegO", "--jpeg-optimize", action="store_true", help="Optimize JPEG file")
    format_group.add_argument("-jpegQ", "--jpeg-quality", metavar="NUM", type=int, help="Quality of JPEG file (0-100)")
    format_group.add_argument("-webpL", "--webp-lossless", action="store_true", help="Save WebP as lossless")
    format_group.add_argument("-webpM", "--webp-method", metavar="NUM", type=int, help="WebP saving speed/quality tradeoff (0-6)")
    format_group.add_argument("-webpQ", "--webp-quality", metavar="NUM", type=int, help="Quality of WebP file (1-100)")

    list_group = parser.add_argument_group("List information")
    list_group.add_argument("-listI", "--list-images", action="store_true", default=False, help="List all compatible scanned images")
    list_group.add_argument("-listF", "--list-file", action="store_true", default=False, help="Save list of compatible scanned images as text file")
    list_group.add_argument("-listT", "--list-tasks", action="store_true", default=False, help="List tasks")

    task_group = parser.add_argument_group("Task handling")
    task_group.add_argument("-addA", "--add-all", action="store_true", default=False, help="Add all compatible images")
    task_group.add_argument("-addID", "--add-id", nargs="+", metavar="NUM", type=int, help="Add compatible images by ID")
    task_group.add_argument("-addN", "--add-new", metavar="NUM", type=int, help="Add compatible images by modified timestamp (newest)")
    task_group.add_argument("-addO", "--add-old", metavar="NUM", type=int, help="Add compatible images by modified timestamp (oldest)")
    task_group.add_argument("-addR", "--add-random", metavar="NUM", type=int, help="Add compatible images randomly")
    task_group.add_argument("-remID", "--remove-id", nargs="+", metavar="NUM", type=int, help="Remove already added compatible images by ID")

    gui_group = parser.add_argument_group("GUI Settings")
    gui_group.add_argument("-fontS", "--font-scale", metavar="NUM", type=float, help="Font scale for GUI")
    gui_group.add_argument("-theme", "--theme", metavar="TEXT", type=str, help="Color theme for FreeSimpleGUI")
    gui_group.add_argument("-viewW", "--view-width", metavar="NUM", type=int, help="Max image width inside GUI")
    gui_group.add_argument("-viewH", "--view-height", metavar="NUM", type=int, help="Max image height inside GUI")

    unknown = None

    try:
        args, unknown = parser.parse_known_args(namespace=p)
    except configargparse.ArgumentError as e:
        logger.error(e)
        logger.info("Use -h or --help to view command-line arguments\n")
        p.cont = False
    except SystemExit:
        p.cont = False

    # Handle typos in arguments
    if unknown:
        for arg in unknown:
            logger.error(f"Unknown argument: {arg}")

        p.cont = False
        logger.info("Use -h or --help to view command-line arguments\n")

    # Handle custom config file
    if p.config_file:
        p.path_config_file = p.config_file

    if p.test_mode:
        p.run_mode = "test mode"
    elif p.preview_mode:
        p.run_mode = "preview mode"
    elif p.slice_mode:
        p.run_mode = "slice mode"
    elif p.count_mode:
        p.run_mode = "count mode"
    else:
        p.run_mode = ""

    return p


#!/usr/bin/env python3

import logging
import random
import cv2 as cv
import imutils as im
import numpy as np
import ruamel.yaml

from io import BytesIO
from PIL import Image, ImageTk, ImageEnhance, UnidentifiedImageError
from pillow_lut import load_cube_file

from .imutils.perspective import four_point_transform

def cv_auto_rotate(img, direction):
    logger = logging.getLogger()
    rot = -90

    if direction == "cw":
        rot = 90

    # Rotate the image if it's width is smaller than it's height
    if img.shape[1] < img.shape[0]:
        img = im.rotate_bound(img, angle=rot)
        logger.debug(f"Auto rotating image 90 degrees {direction}")

    return img

# Slice contour area from image
def cv_slice_img(img, img_resized, cnt, pfix):
    logger = logging.getLogger()
    out_slice = None
    rot_rect = None
    slice_with_pfix = False

    x_fac = img.shape[1] / img_resized.shape[1]
    y_fac = img.shape[0] / img_resized.shape[0]
    cnt[:,:,0] = cnt[:,:,0] * x_fac
    cnt[:,:,1] = cnt[:,:,1] * y_fac

    # Do we need perspective fix?
    if pfix != 0:
        rot_rect = cv.minAreaRect(cnt)

        # Calculate tilt angle
        tilt_angle = abs(rot_rect[2])

        # Enable perspective fix if the tilt angle is larger than the maximum allowed tilt angle
        if tilt_angle > pfix and tilt_angle < (90 - pfix):

            if tilt_angle < 45.0:
                logger.debug(f"Perspective fix image for {tilt_angle} degree tilt")
            else:
                logger.debug(f"Perspective fix image for {90 - tilt_angle} degree tilt")

            slice_with_pfix = True

    # Slice image with pfix
    if slice_with_pfix:
        pts = np.int64(cv.boxPoints(rot_rect))
        out_slice = four_point_transform(img, pts, (255, 255, 255))

    # Slice image without pfix
    else:
        x, y, w, h = cv.boundingRect(cnt)

        out_slice = img[
            y:min(y + h, img.shape[0]),
            x:min(x + w, img.shape[1])
            ]

    return out_slice

# Check if cnt is between range(min, max)
def cv_is_cnt_in_range(img, cnt, min, max):
    img_area = img.data.shape[0] * img.shape[1]
    x, y, w, h = cv.boundingRect(cnt)
    cnt_area = w * h
    cnt_area_in_prc = cnt_area / img_area * 100.0

    if cnt_area_in_prc > min and cnt_area_in_prc < max:
        return True

    return False

def cv_draw_rect(img, rect, color, border=2):
    logger = logging.getLogger()
    cv.rectangle(img, (rect.x, rect.y), (rect.x + rect.w, rect.y + rect.h), color, border)
    logger.debug(f"Draw rectangle ({rect.w}x{rect.h}) with color {color}")

def cv_draw_cnt(img, cnt, color, border=2):
    logger = logging.getLogger()
    x, y, w, h = cv.boundingRect(cnt)
    cv.rectangle(img, (x, y), (x + w, y + h), color, border)
    logger.debug(f"Draw rectangle ({w}x{h}) with color {color}")

# Detect image slices and return them as contours
def cv_detect_slices(img):
    cnts = cv.findContours(img, cv.RETR_EXTERNAL, cv.CHAIN_APPROX_SIMPLE)
    return im.grab_contours(cnts)

# Apply white threshold to OpenCV image
def cv_apply_wt(img, wt):
    logger = logging.getLogger()
    img_wt_gray = cv.cvtColor(img, cv.COLOR_BGR2GRAY)
    img_wt_blur = cv.GaussianBlur(img_wt_gray, (5, 5), 0)
    img_wt_thresh = cv.threshold(img_wt_blur, wt, 255, cv.THRESH_BINARY_INV)[1]
    logger.debug(f"Apply threshold with value {wt}")
    return img_wt_thresh

def cv_to_pil(img):
    return Image.fromarray(cv.cvtColor(img, cv.COLOR_BGR2RGB))

def cv_denoise(img, strength):
    logger = logging.getLogger()
    strength = int(strength)
    param = [3, 3]

    match strength:
        case 0: return
        case 1: param.extend([3, 3])
        case 2: param.extend([5, 7])
        case 3: param.extend([7, 11])
        case 4: param.extend([9, 15])
        case 5: param.extend([11, 19])

    img_denoise = cv.fastNlMeansDenoisingColored(
        img,
        None,
        param[0],
        param[1],
        param[2],
        param[3]
        )

    logger.debug(f"Apply denoise with params {param}")

    return img_denoise

def cv_resize(img, w=None, h=None, scale=None):
    logger = logging.getLogger()

    if scale:
        logger.debug(f"CV Scale image: ({img.shape[1]}x{img.shape[0]}) -> ({scale * img.shape[1]}x{scale * img.shape[0]})")

        return cv.resize(img, None, fx=scale, fy=scale, interpolation=cv.INTER_AREA)

    if w:
        logger.debug(f"CV Resize image width ({img.shape[1]}) -> ({w})")

        return im.resize(img, width=w)

    if h:
        logger.debug(f"CV Resize image height ({img.shape[0]}) -> ({h})")

        return im.resize(img, height=h)

def pil_resize(img, w=None, h=None, scale=None):
    logger = logging.getLogger()

    if scale:
        logger.debug(f"PIL Scale image {img.size} -> {scale * img.size}")

        new_w, new_h = (int(img.width * scale), int(img.height * scale))
        return img.resize((new_w, new_h))

    if w:
        ratio = w / img.width

        if w != img.width:
            logger.debug(f"PIL Resize image width ({img.width}) -> ({w})")
            return img.resize((w, int(img.height * ratio)))
        else:
            return img

    if h:
        ratio = h / img.height

        if h != img.height:
            logger.debug(f"PIL Resize image height ({img.height}) -> ({h})")
            return img.resize((int(img.width * ratio), h))
        else:
            return img

def pil_open_image(filepath):
    logger = logging.getLogger()
    img = None

    try:
        img = Image.open(filepath)
    except (FileNotFoundError, UnidentifiedImageError) as e:
        logger.error(e)

    return img

def pil_to_cv(img):
    return cv.cvtColor(np.array(img), cv.COLOR_RGB2BGR)

def pil_to_buffer(img):
    with BytesIO() as buf:
        img.save(buf, format="PNG")
        output = buf.getvalue()

    return output

def pil_filter_image(img, filters):
    logger = logging.getLogger()
    color = filters[0]
    contrast = filters[1]
    brightness = filters[2]
    sharpness = filters[3]
    denoise = filters[4]
    lut_str = filters[5]
    lut_path = filters[6]

    if denoise:
        img_filter = cv_to_pil(cv_denoise(img, denoise))
    else:
        img_filter = cv_to_pil(img)

    if lut_path and lut_str > 0.0:
        lut = load_cube_file(lut_path)
        img_filter_lut = img_filter.filter(lut)
        img_filter = Image.blend(img_filter, img_filter_lut, lut_str)

        logger.debug(f"Filter image with LUT using strength: {lut_str}")

    if color != 1.0:
        enhanced_img = ImageEnhance.Color(img_filter)
        img_filter = enhanced_img.enhance(color)
        logger.debug(f"Filter color with value: {color}")

    if contrast != 1.0:
        enhanced_img = ImageEnhance.Contrast(img_filter)
        img_filter = enhanced_img.enhance(contrast)
        logger.debug(f"Filter contrast with value: {contrast}")

    if brightness != 1.0:
        enhanced_img = ImageEnhance.Brightness(img_filter)
        img_filter = enhanced_img.enhance(brightness)
        logger.debug(f"Filter brightness with value: {brightness}")

    if sharpness != 1.0:
        enhanced_img = ImageEnhance.Sharpness(img_filter)
        img_filter = enhanced_img.enhance(sharpness)
        logger.debug(f"Filter sharpness with value: {sharpness}")

    return img_filter

def random_string(rand1, rand2, rand3):
    random.seed(rand1 + rand2 + rand3)
    return str(random.random()).replace("0.", "")

def remove_suffix(this):
    count = 0
    first_dot_index = 0

    for char in reversed(this):
        if char == ".":
            first_dot_index = count
            break
        else:
            count += 1

    return this[:-first_dot_index - 1]

def convert_bytes(size):
    power = 2**10
    n = 0
    labels = {0: '', 1: ' Kb', 2: ' Mb', 3: ' Gb', 4: ' Tb'}

    while size > power:
        size /= power
        n += 1

    return str(round(size)) + str(labels[n])

def yaml_change_value(path, yaml_key, new_val):
    with open(path, 'r') as f:
        yaml = ruamel.yaml.YAML()
        yaml.preserve_quotes = True
        data = yaml.load(f)

    data[yaml_key] = new_val

    with open(path, 'w') as f:
        yaml.dump(data, f)

def create_statusline(name, version, description, config):
    desc_words = description.split(" ")

    art_lines = [
        "   ___________   ",
        "  |  __   __  |  %s - version: %s" % (name, version),
        "  | |S | |C | |  ",
        "  | |__| |__| |  %s " % " ".join(desc_words[:6]).strip("\n"),
        "  |  __   __  |  %s" % " ".join(desc_words[6:]).strip("\n"),
        "  | |I | |S | |  ",
        "  | |__| |__| |  %s" % "Config file:",
        "  |___________|  %s" % config,
    ]

    art = "\n".join(art_lines)
    art += "\n"

    return art



#!/usr/bin/env python3

import os
import timeit
import logging
import FreeSimpleGUI as sg

from PIL import Image
from .utils import *
from .gui_widgets import *

# Popup info message
def popup_msg(msg):
    sg.popup(f"{msg}", auto_close=True, auto_close_duration=3, no_titlebar=True)

# Show preview of sliced image
def show_preview_gui(p, scis_img):
    logger = logging.getLogger()
    sg.theme(p.theme)
    window_title = "Scan-Image-Slicer - Preview mode"
    layout = []
    widgets = []
    preview_images = scis_img.create_preview_images(p)
    image_count = len(preview_images)
    sentinel = False
    image_index = 0
    filter_time = 0
    unapplied_filters = []

    applied_filters = [
        p.filter_color,
        p.filter_contrast,
        p.filter_brightness,
        p.filter_sharpness,
        p.filter_denoise,
        p.filter_lut_strength,
        p.filter_lut_path
    ]

    # Output warning if no images found and skip the file
    if not preview_images:
        logger.warning(f"[ID:{scis_img.id}] - ({scis_img.name}) - No images found, skipping it..")
        return 0, sentinel

    # Create menu strings
    def str_slice_size():
        img = preview_images[image_index]
        return f"Size: {img.shape[0]}x{img.shape[1]} px"

    def str_slice_n():
        return f"Slice {image_index+1} of {image_count}"

    def str_filter_time():
        return f"Filter time: {filter_time} seconds"

    # Use LUT?
    if p.filter_lut_path:
        use_lut = True
    else:
        use_lut = False

    # Info section
    widgets.append([Header(p, scis_img.name)])
    widgets.append([Text(p, str_slice_n(), key="slice_n")])
    widgets.append([Text(p, str_slice_size(), key="slice_size")])
    widgets.append([Text(p, str_filter_time(), key="filter_time", visible=False)])
    widgets.append([sg.HorizontalSeparator(p=(0, 5))])

    # Filters section
    widgets.append([Header(p, "Denoise:")])
    widgets.append([Slider(p, "int", (0, 5), p.filter_denoise, key="filter_denoise")])
    widgets.append([Header(p, "LUT:", visible=use_lut)])
    widgets.append([Slider(p, "float", (0.0, 1.0), p.filter_lut_strength, key="filter_lut_strength", visible=use_lut)])
    widgets.append([Header(p, "Color:")])
    widgets.append([Slider(p, "float", (0.0, 2.0), p.filter_color, key="filter_color")])
    widgets.append([Header(p, "Contrast:")])
    widgets.append([Slider(p, "float", (0.0, 2.0), p.filter_contrast, key="filter_contrast")])
    widgets.append([Header(p, "Brightness:")])
    widgets.append([Slider(p, "float", (0.0, 2.0), p.filter_brightness, key="filter_brightness")])
    widgets.append([Header(p, "Sharpness:")])
    widgets.append([Slider(p, "float", (0.0, 2.0), p.filter_sharpness, key="filter_sharpness")])
    widgets.append([Text(p, "")])
    widgets.append([Button(p, "Apply filters", key="apply_filters", disabled=True), Button(p, "Save filters", key="save_filters")])
    widgets.append([sg.HorizontalSeparator(p=(0, 5))])

    # Actions section
    widgets.append([Header(p, "Actions:")])
    widgets.append([Button(p, "Save image"), Button(p, "Open image")])

    if image_count > 1:
        widgets.append([Button(p, "Last slice", disabled=True), Button(p, "Next slice")])

    widgets.append([Button(p, "Abort"), Button(p, "Next image")])

    # Apply filters on sliced image
    filter_image = pil_filter_image(preview_images[image_index], applied_filters)

    # Resize filtered image to fit GUI
    filter_image = pil_resize(filter_image, h=min(filter_image.height, p.view_height))
    filter_image = pil_resize(filter_image, w=min(filter_image.width, p.view_width))

    layout.append([Image(pil_to_buffer(filter_image), key="preview_image"), Column(widgets)])

    # Create the GUI window
    window = Window(window_title, layout)

    # Run the GUI
    while True:
        event, values = window.read()

        # Abort run
        if event in [sg.WIN_CLOSED, "Abort"]:
            logger.info("Aborting")
            sentinel = True
            break

        # Next image
        if event == "Next image":
            break

        # Update filter values on slider events
        if event in ["filter_color", "filter_contrast", "filter_brightness", "filter_sharpness", "filter_denoise", "filter_lut_strength"]:
            unapplied_filters = [
                values["filter_color"],
                values["filter_contrast"],
                values["filter_brightness"],
                values["filter_sharpness"],
                values["filter_denoise"],
                values["filter_lut_strength"],
                p.filter_lut_path
            ]

            if unapplied_filters != applied_filters:
                window["apply_filters"].update(disabled=False)
            else:
                window["apply_filters"].update(disabled=True)

        # Apply filters to image
        if event == "apply_filters":

            p.filter_color = values["filter_color"]
            p.filter_contrast = values["filter_contrast"]
            p.filter_brightness = values["filter_brightness"]
            p.filter_sharpness = values["filter_sharpness"]
            p.filter_denoise = int(values["filter_denoise"])
            p.filter_lut_strength = values["filter_lut_strength"]

            # Time filters
            start = timeit.default_timer()
            filter_image = pil_filter_image(preview_images[image_index], unapplied_filters)
            stop = timeit.default_timer()
            filter_time = round(stop - start, 4)

            # Resize filtered image to fit GUI
            filter_image = pil_resize(filter_image, h=min(filter_image.height, p.view_height))
            filter_image = pil_resize(filter_image, w=min(filter_image.width, p.view_width))

            window["preview_image"].update(pil_to_buffer(filter_image))

            window["apply_filters"].update(disabled=True)
            window["filter_time"].update(str_filter_time(), visible=True)
            applied_filters = unapplied_filters

        # Save filter values to config file
        if event == "save_filters":
            yaml_change_value(p.path_config_file, "filter-color", values["filter_color"])
            yaml_change_value(p.path_config_file, "filter-contrast", values["filter_contrast"])
            yaml_change_value(p.path_config_file, "filter-brightness", values["filter_brightness"])
            yaml_change_value(p.path_config_file, "filter-sharpness", values["filter_sharpness"])
            yaml_change_value(p.path_config_file, "filter-denoise", int(values["filter_denoise"]))
            yaml_change_value(p.path_config_file, "filter-lut-strength", values["filter_lut_strength"])
            popup_msg(f"New filter values saved to:\n {p.path_config_file}")

        # Open image with OS default handler
        if event == "Open image":
            img = pil_filter_image(preview_images[image_index], applied_filters)
            img.show()

        # Save image to desired location
        if event == "Save image":

            # Get file path where to save image
            filepath = sg.popup_get_file(
                "Save image",
                default_path=os.path.join(
                    p.output,
                    f"slice_{image_index+1}_from_" + scis_img.name
                ),
                save_as=True,
            )

            # Try to save the file
            if filepath:
                img = pil_filter_image(preview_images[image_index], applied_filters)
                img.save(filepath)
                popup_msg(f"Saved image to:\n {filepath}")

        # Load next slice
        if event == "Next slice":
            image_index += 1

            # Handle navigation
            if image_index in range(0, image_count-1):
                window["Next slice"].update(disabled=False)
            else:
                window["Next slice"].update(disabled=True)

            if image_index == 1:
                window["Last slice"].update(disabled=False)

        # Load last image
        if event == "Last slice":
            image_index -= 1

            # Handle navigation
            if image_index in range(1, image_count):
                window["Last slice"].update(disabled=False)
            else:
                window["Last slice"].update(disabled=True)

            if image_index == image_count-2:
                window["Next slice"].update(disabled=False)

        if event in ["Next slice", "Last slice"]:

            if window["apply_filters"].Disabled:
                applied_filters = [
                    p.filter_color,
                    p.filter_contrast,
                    p.filter_brightness,
                    p.filter_sharpness,
                    p.filter_denoise,
                    p.filter_lut_strength,
                    p.filter_lut_path
                ]

            # Update data
            window["slice_n"].update(str_slice_n())
            window["slice_size"].update(str_slice_size())

            # Time filter
            start = timeit.default_timer()
            filter_image = pil_filter_image(preview_images[image_index], applied_filters)
            stop = timeit.default_timer()
            filter_time = round(stop - start, 4)

            # Resize filtered image to fit GUI
            filter_image = pil_resize(filter_image, h=min(filter_image.height, p.view_height))
            filter_image = pil_resize(filter_image, w=min(filter_image.width, p.view_width))

            window["preview_image"].update(pil_to_buffer(filter_image))
            window["filter_time"].update(str_filter_time(), visible=True)

    window.close()
    return image_count, sentinel

# Show test image inside GUI
def show_test_gui(p, scis_img):
    logger = logging.getLogger()
    sg.theme(p.theme)
    window_title = "Scan-Image-Slicer - Test mode"
    layout = []
    widgets = []
    test_image = scis_img.create_test_image(p)
    sentinel = False

    # Create info strings
    def update_detection_count_txt(count):
        return f"Detected images (blue): {count}"

    def update_ignored_count_txt(count):
        return f"Detected false images (red): {count}"

    # Create layout for GUI widgets
    widgets.append([Header(p, scis_img.name)])
    widgets.append([Text(p, update_detection_count_txt(scis_img.slice_count), key="slice_count")])
    widgets.append([Text(p, update_ignored_count_txt(scis_img.false_slice_count), key="false_slice_count")])
    widgets.append([sg.HorizontalSeparator(p=(0, 5))])
    widgets.append([Header(p, "Image detection sensitivity:")])
    widgets.append([Slider(p, "int", (0, 255), p.white_threshold, key="white_threshold")])
    widgets.append([Text(p, "Hint: try values between 220-250")])
    widgets.append([Header(p, "Minimum slice size in %:")])
    widgets.append([Slider(p, "float", (0.01, 99.99), p.minimum_size, resolution=0.01, key="minimum_size")])
    widgets.append([Header(p, "Maximum slice size in %:")])
    widgets.append([Slider(p, "float", (0.01, 99.99), p.maximum_size, resolution=0.01, key="maximum_size")])
    widgets.append([sg.Text("", p=(0, 2))])
    widgets.append([Button(p, "Detect images", disabled=False), Button(p, "Save values")])
    widgets.append([sg.HorizontalSeparator(p=(0, 5))])
    widgets.append([Header(p, "Actions:")])
    widgets.append([Button(p, "Save image"), Button(p, "Open image")])
    widgets.append([Button(p, "Abort"), Button(p, "Next image")])

    # Create layout for window
    layout.append([Image(pil_to_buffer(test_image), key="test_image"), Column(widgets)])

    # Create window
    window = Window(window_title, layout)

    # Run the GUI
    while True:
        event, values = window.read()

        # Abort run
        if event in [sg.WIN_CLOSED, "Abort"]:
            logger.info("Aborting")
            sentinel = True
            break

        # Load next image
        if event == "Next image":
            break

        # Open image with OS default handler
        if event == "Open image":
            test_image.show()

        # Save image to desired location
        if event == "Save image":

            # Get file path where to save image
            filepath = sg.popup_get_file("Save image", default_path=os.path.join(p.output, scis_img.name), save_as=True)

            # Try to save the file
            if filepath:
                test_image.save(filepath)
                popup_msg(f"Saved image to:\n {filepath}")

        # Save values
        if event == "Save values":
            yaml_change_value(p.path_config_file, "white-threshold", int(values["white_threshold"]))
            yaml_change_value(p.path_config_file, "minimum-size", values["minimum_size"])
            yaml_change_value(p.path_config_file, "maximum-size", values["maximum_size"])
            p.white_threshold = values["white_threshold"]
            p.minimum_size = values["minimum_size"]
            p.maximum_size = values["maximum_size"]
            popup_msg(f"Values saved to:\n {p.path_config_file}")

        # Detect images
        if event == "Detect images":
            scis_img.slice_count = 0
            scis_img.false_slice_count = 0
            p.white_threshold = values["white_threshold"]
            p.minimum_size = values["minimum_size"]
            p.maximum_size = values["maximum_size"]
            test_image = scis_img.create_test_image(p)
            window["test_image"].update(pil_to_buffer(test_image))
            window["slice_count"].update(update_detection_count_txt(scis_img.slice_count))
            window["false_slice_count"].update(update_ignored_count_txt(scis_img.false_slice_count))

    window.close()
    return scis_img.slice_count, sentinel


#!/usr/bin/env python3

import os
import timeit
import random
import logging

from datetime import datetime
from concurrent.futures import ProcessPoolExecutor, as_completed
from operator import itemgetter
from tqdm.auto import tqdm
from time import strftime, localtime, gmtime
from .gui import show_preview_gui, show_test_gui
from .scis_image import ScanImageSlicerImage
from .utils import *

def run_tasks(queue, p, tasks, images):
    logger = logging.getLogger()

    # Create unique run id
    p.run_id = p.project_name + "_{:%Y_%m_%d_%H_%M_%S}".format(datetime.now())

    # Create unique run path
    p.unique_path = os.path.join(p.output, p.run_id)

    result = 0
    workers = 1

    start = timeit.default_timer()
    logger.info("%s started @ %s", p.run_mode.capitalize(), strftime("%a %d %b %Y %H:%M:%S", localtime()))

    # Enable more workers if needed
    if p.count_mode or p.slice_mode:
        workers = p.workers

    # Do we need multiprocessing?
    if workers > 1 and len(tasks) > 1:

        logger.info(f"Use multiprocessing with {workers} workers")

        with ProcessPoolExecutor(max_workers=workers) as executor:

            # Create progress bar for our tasks
            with tqdm(total=len(tasks), desc=":: Progress", unit=" images", ncols=100) as pbar:
                futures = {}

                # Split our tasks for the executor
                for i, task in enumerate(tasks):

                    if p.count_mode:
                        future = executor.submit(images[task].count_slices, queue, p)
                    elif p.slice_mode:
                        future = executor.submit(images[task].save_slices, queue, p)

                    futures[future] = i

                for future in as_completed(futures):
                    i = futures[future]
                    result += future.result()
                    pbar.update(1)
    else:

        # Create progress bar for our tasks
        pbar = tqdm(tasks, desc=":: Progress", unit=" images", ncols=100)

        # Go over tasks one by one
        for task in pbar:
            if p.count_mode:
                result += images[task].count_slices(queue, p)

            if p.test_mode:
                val, sentinel = show_test_gui(p, images[task])
                result += val

                if sentinel:
                    break

            elif p.preview_mode:
                val, sentinel = show_preview_gui(p, images[task])
                result += val

                if sentinel:
                    break

            elif p.slice_mode:
                result += images[task].save_slices(queue, p)

    # Stop timer and calculate time lapsed
    stop = timeit.default_timer()
    seconds = (stop - start)
    timer_result = strftime("%H hours, %M minutes and %S seconds", gmtime(seconds))

    logger.info("%s finished @ %s", p.run_mode.capitalize(), strftime("%a %d %b %Y %H:%M:%S", localtime()))
    logger.info("Time lapsed: %s", timer_result)

    # Report results
    action = "Sliced" if p.slice_mode else "Detected"
    logger.info(f"{action} {result} images\n")

    if p.slice_mode:
        logger.info(f"Output: {p.unique_path}\n")

# Parse parameters
def parse_p(p):
    logger = logging.getLogger()
    errors = []

    # Sanitize value paths
    p.input = os.path.normpath(os.path.expanduser(p.input))
    p.output = os.path.normpath(os.path.expanduser(p.output))

    if p.filter_lut_path:
        p.filter_lut_path = os.path.normpath(os.path.expanduser(p.filter_lut_path))

    # Make sure we have some project name
    if not p.project_name:
        p.project_name = "NewProject"

    # Make sure input path exists
    if not os.path.exists(p.input):
        errors.append(f"Input path does not exist: {p.input}")

    # Make sure output path exists
    if not os.path.exists(p.output):
        errors.append(f"Output path does not exist: {p.output}")

    # Make sure we have valid LUT file
    if p.filter_lut_path:
        if os.path.isfile(p.filter_lut_path):
            if not p.filter_lut_path.lower().endswith(".cube"):
                errors.append(f"LUT should be a valid .cube file")
        else:
            errors.append(f"Could not find LUT file at: {p.filter_lut_path}")

    # Go over value ranges
    if not p.white_threshold in range(1, 256):
        errors.append("Value of '-white/--white-threshold' should be between 1 and 255")

    if not int(p.minimum_size) in range(1, 101):
        errors.append("Value of '-min/--minimum-size' should be between 1 and 100")

    if not int(p.maximum_size) in range(1, 101):
        errors.append("Value of '-max/--maximum-size' should be between 1 and 100")

    if not p.view_height >= 100:
        errors.append("Value of '-viewH/--view_height' should be at least 100")

    if not p.view_width >= 100:
        errors.append("Value of '-viewW/--view_width' should be at least 100")

    if not p.filter_denoise in range(0, 6):
        errors.append("Value of '-denoise/--filter-denoise' should be between 0 and 5")

    if not int(p.filter_color * 100.0) in range(0, 201):
        errors.append("Value of '-color/--filter-color' should be between 0.0 and 2.0")

    if not int(p.filter_contrast * 100.0) in range(0, 201):
        errors.append("Value of '-contrast/--filter-contrast' should be between 0.0 and 2.0")

    if not int(p.filter_brightness * 100.0) in range(0, 201):
        errors.append("Value of '-brightness/--filter-brightness' should be between 0.0 and 2.0")

    if not int(p.filter_sharpness * 100.0) in range(0, 201):
        errors.append("Value of '-sharpness/--filter-sharpness' should be between 0.0 and 2.0")

    if not int(p.filter_lut_strength * 100.0) in range(0, 101):
        errors.append("Value of '-lutS/--filter-lut-strength' should be between 0.0 and 1.0")

    if not p.perspective_fix in range(0, 90):
        errors.append("Value of '-pfix/--perspective-fix' should be between 0 and 89")

    if not p.auto_rotate in ["disable", "cw", "ccw"]:
        errors.append("Value of '-autoR/--auto-rotate' should be one of disable, cw or ccw")

    if not p.save_format in ["jpeg", "png", "webp"]:
        errors.append("Value of '-save/--save-format' should be one of jpeg, png or webp")

    if not p.png_compression in range(0, 10):
        errors.append("Value of '-pngC/--png-compression' should be between 0 and 9")

    if not p.jpeg_quality in range(0, 96):
        errors.append("Value of '-jpegQ/--jpeg-quality' should be between 0 and 95")

    if not p.webp_method in range(0, 7):
        errors.append("Value of '-webpM/--webp-method' should be between 0 and 6")

    if not p.webp_quality in range(1, 101):
        errors.append("Value of '-webpQ/--webp-quality' should be between 1 and 100")

    # Abort on errors
    if errors:
        logger.info("Fix the following errors to continue:")
        for err in errors:
            logger.error(err)
        p.cont = False

    return p

# Confirm the function
def confirm(skip, task_count, run_mode):
    logger = logging.getLogger()
    logger.info(f"Starting {run_mode} with {task_count} tasks")

    if skip:
        return True

    while True:
        try:
            logger.info("Continue? y/n")
            answer = input("").lower()

            if answer == 'y':
                return True
            elif answer == 'n':
                logger.info("Aborting\n")
                return False
            else:
                continue
        except (OSError, EOFError) as e:
            logger.error(e)
            return False
        except KeyboardInterrupt:
            return

def create_dots(chars):
    cutoff_point = 6
    cutoff_amount = chars - cutoff_point
    dots = " .......... "
    dots = dots.replace('.', "", cutoff_amount)

    return dots

def list_images(images):
    logger = logging.getLogger()
    logger.info("List images:")

    for key in images.keys():
        line = "[ID:" + str(images[key].id) + "]"
        line += create_dots(len(line)) + images[key].name
        line += " (" + images[key].size_mb + ")"
        logger.info(line)

    logger.info(f"Images found: {len(images)}\n")

# Save list of images as a txt file inside output directory
def save_imagelist_as_txt(path, name, images):
    logger = logging.getLogger()
    fn = f"{name}_images.txt"
    fp = os.path.join(path, fn)
    lines = []

    for key in images.keys():
        line = "[ID:" + str(images[key].id) + "]"
        line += create_dots(len(line)) + images[key].name
        line += " (" + images[key].size_mb + ")"
        lines.append(line + "\n")

    with open(fp, 'w') as outfile:
        try:
            outfile.writelines(lines)
            outfile.close()
        except OSError as e:
            logger.error(e + "\n")

    if os.path.exists(fp):
        logger.info("Saved list of images:")
        logger.info(fp + "\n")

def handle_tasks(p, images):
    logger = logging.getLogger()
    tasks = []
    warns = []
    info = []

    def add_valid_task(task_id):

        if not task_id in tasks:
            if task_id in images:
                tasks.append(task_id)
            else:
                warns.append(f"Image not found with ID: {task_id}")
        else:
            warns.append(f"Task already created for image with ID: {task_id}")

    def remove_valid_task(task_id):

        if task_id in tasks:
            tasks.remove(task_id)
        else:
            warns.append(f"No task for image with ID: {task_id}")

    # Add all images
    if p.add_all:

        info.append(f"Create task for all {len(images)} images")

        tasks = list(images.keys())

    # Add images with ID
    if p.add_id:

        info.append(f"Create task for images with ID: {p.add_id}")

        for i in p.add_id:
            add_valid_task(i)

    # Add new images by mtime
    if p.add_new:

        info.append(f"Create task for {p.add_new} newest images by modification date")

        for i in range(p.add_new):
            add_valid_task(i)

    # Add old images by mtime
    if p.add_old:

        info.append(f"Create task for {p.add_old} oldest images by modification date")

        for i in range(p.add_old):
            add_valid_task((len(images)-1)-i)

    # Add random images
    if p.add_random:

        pool = list(images.keys())
        picks = []

        for i in range(p.add_random):
            if pool:
                pick = random.choice(pool)

                if not pick in picks:
                    picks.append(pick)
                    pool.remove(pick)

        info.append(f"Create task for random images with ID: {picks}")

        for pick in picks:
            add_valid_task(pick)

    # Remove tasks by ID
    if p.remove_id:

        info.append(f"Remove task for image with ID: {p.remove_id}")

        for i in p.remove_id:
            remove_valid_task(i)

    if info:
        for i in info:
            logger.info(i)

    if warns:
        for warn in warns:
            logger.warning(warn)

    logger.info(f"Tasks added: {len(tasks)}\n")

    return tasks

def list_tasks(tasks, images):
    logger = logging.getLogger()
    logger.info("List tasks:")

    if tasks:
        for task in tasks:
            image = images[task]
            line = "[ID:" + str(image.id) + "]"
            line += create_dots(len(line)) + image.name
            line += " (" + image.size_mb + ")"
            logger.info(line)

        logger.info(f"Total: {len(tasks)}\n")
    else:
        logger.info("Tasklist is empty\n")

def sequential_parallel_rename(p):
    logger = logging.getLogger()
    output_images = []
    counter = 0
    cur_path = ""
    result = 0

    if p.save_format == "png":
        suffix = ".png"
    if p.save_format == "jpeg":
        suffix = ".jpg"
    if p.save_format == "webp":
        suffix = ".webp"

    # Walk through the unique directory inside output path and collect temporary images
    for path, dirs, files in os.walk(os.path.normpath(p.unique_path)):

        # Go through files one by one and rename them based on the directory names
        for file in files:

            # Only handle files with the right name and format
            if file.lower().startswith("tmp_file_") and file.lower().endswith(suffix):

                # Reset counter if our path changes
                if cur_path != path:
                    counter = 0
                    cur_path = path

                counter += 1

                # Get relative path of file
                rel_path = os.path.relpath(path, p.unique_path)

                # Handle files that are in the root dir (.)
                if rel_path == ".":
                    image = [
                        os.path.join(path, file),
                        os.path.join(path, "output_" + str(counter)) + suffix
                    ]

                    output_images.append(image)

                # Handle files within folders
                else:
                    filename = ""
                    parts = rel_path.split(os.sep)

                    for part in parts:
                        filename += part + "_"

                    image = [
                        os.path.join(path, file),
                        os.path.join(path, filename + str(counter) + suffix)
                    ]

                    output_images.append(image)

    # Do we need multiprocessing?
    if p.workers > 1 and len(output_images) > 1:
        with ProcessPoolExecutor(max_workers=p.workers) as executor:

            # Split our output images for the executor
            for i, output_image in enumerate(output_images):
                executor.submit(os.rename, output_image[0], output_image[1])
                logger.debug(f"Rename {output_image[0]} to {output_image[1]}")
                result += i
    else:
        for output_image in output_images:
            os.rename(output_image[0], output_image[1])
            logger.debug(f"Rename {output_image[0]} to {output_image[1]}")

# Collect images from input directory
def collect_images(input):

    id = 0
    sorted_images = {}
    scanned_images = []

    # Only accept these formats for the Image
    accepted_formats = [
                ".bmp",
                ".jpeg",
                ".jpg",
                ".png",
                ".webp",
                ".tiff",
            ]

    # Walk through the input folder and collect images
    for path, dirs, files in os.walk(os.path.normpath(input)):

        # Go through files one by one
        for file in files:
            skip_file = True
            format = ""

            # Only accept image formats defined above
            for suffix in accepted_formats:
                if file.lower().endswith(suffix):
                    format = suffix[1:]
                    skip_file = False

            if not skip_file:

                # Append our found data
                image = [
                    path,
                    file,
                    format,
                    os.stat(os.path.join(path, file)).st_mtime,
                    os.stat(os.path.join(path, file)).st_size]

                scanned_images.append(image)

    if scanned_images:

        # Sort our images from newest to oldest (mtime)
        scanned_images = sorted(scanned_images, key=itemgetter(3), reverse=True)

        for scanned_image in scanned_images:

                image = ScanImageSlicerImage(
                    id,
                    scanned_image[0],
                    scanned_image[1],
                    scanned_image[2],
                    scanned_image[3],
                    scanned_image[4],
                )

                # Add our image to dictionary using the image id as key
                sorted_images[image.id] = image
                id += 1

        # Return our sorted images as a dictionary
        return sorted_images


#!/usr/bin/env python3

import os
import sys
import logging
import logging.handlers

LOGGING_LEVEL = logging.INFO

def listener_configurer(p):
    logger = logging.getLogger()
    stream_handler = logging.StreamHandler()
    file_handler = logging.FileHandler(p.log_path, mode='w')
    stream_formatter = logging.Formatter(":: " + "%(message)s")

    file_formatter = logging.Formatter(
        "[%(asctime)s] [%(levelname)s] - %(message)s",
        datefmt="%d %b %Y %a %I:%M:%S %p"
    )

    stream_handler.setFormatter(stream_formatter)
    file_handler.setFormatter(file_formatter)

    logger.addHandler(stream_handler)
    logger.addHandler(file_handler)

def listener_process(p, queue, configurer):
    configurer(p)
    while True:
        try:
            record = queue.get()
            if record is None:
                break
            logger = logging.getLogger(record.name)
            logger.handle(record)
        except Exception:
            import sys, trackeback
            print("Logger error: ", file=sys.stderr)
            trackeback.print_exc(file=sys.stderr)

def queue_configurer(queue):
    queue_handler = logging.handlers.QueueHandler(queue)
    logger = logging.getLogger()

    if not logger.handlers:
        logger.addHandler(queue_handler)

    logger.setLevel(LOGGING_LEVEL)

    return logger

def worker_configurer(queue):
    queue_handler = logging.handlers.QueueHandler(queue)
    logger = logging.getLogger()
    logger.addHandler(queue_handler)
    logger.setLevel(LOGGING_LEVEL)


#!/usr/bin/env python3

import os
from .utils import *
from .scis_logger import queue_configurer

class ScanImageSlicerImage:
    def __init__(self, id, path, name, format, mtime, size):
        self.id = id
        self.path = path
        self.name = name
        self.format = format
        self.mtime = mtime
        self.size = size
        self.size_mb = convert_bytes(size)
        self.filepath = os.path.join(path, name)
        self.slice_count = 0
        self.false_slice_count = 0

    # Count slices inside scanned image
    def count_slices(self, queue, p):
        logger = queue_configurer(queue)
        img = pil_open_image(self.filepath)

        if not img:
            return 0

        img = pil_to_cv(img)
        img_resized = cv_resize(img, w=min(900, img.shape[1]))

        for cnt in cv_detect_slices(cv_apply_wt(img_resized, p.white_threshold)):
            if cv_is_cnt_in_range(img_resized, cnt, p.minimum_size, p.maximum_size):
                self.slice_count += 1

        # Output warning if no images found
        if not self.slice_count:
            logger.warning(f"[ID:{self.id}] - ({self.name}) - No images found, skipping it..")
            return 0

        return self.slice_count

    # Slice images and save them to the output folder
    def save_slices(self, queue, p):
        logger = queue_configurer(queue)
        img = pil_open_image(self.filepath)

        if not img:
            return 0

        img = pil_to_cv(img)
        img_resized = cv_resize(img, w=min(900, img.shape[1]))

        # Define file format settings
        file_params = {}
        file_params.setdefault("format", p.save_format)

        if p.save_format == "png":
            savefile_suffix = ".png"
            file_params.setdefault("optimize", p.png_optimize)
            file_params.setdefault("compress_level", p.png_compression)
        if p.save_format == "jpeg":
            savefile_suffix = ".jpg"
            file_params.setdefault("optimize", p.jpeg_optimize)
            file_params.setdefault("quality", p.jpeg_quality)
        if p.save_format == "webp":
            savefile_suffix = ".webp"
            file_params.setdefault("lossless", p.webp_lossless)
            file_params.setdefault("method", p.webp_method)
            file_params.setdefault("quality", p.webp_quality)

        # Define filters
        filters = [
            p.filter_color,
            p.filter_contrast,
            p.filter_brightness,
            p.filter_sharpness,
            p.filter_denoise,
            p.filter_lut_strength,
            p.filter_lut_path
        ]

        # Define save path
        save_path = os.path.normpath(os.path.join(p.unique_path, os.path.relpath(self.path, p.input)))

        # Create save path
        if not os.path.exists(save_path):
            os.makedirs(save_path, exist_ok=True)

        # Make sure path was created
        if not os.path.exists(save_path):
            logger.error(f"Could not create directory: {save_path}")
            return 0

        # Loop through cnts and save slices
        for cnt in cv_detect_slices(cv_apply_wt(img_resized, p.white_threshold)):

            # Make sure the area is between min/max sizes
            if cv_is_cnt_in_range(img_resized, cnt, p.minimum_size, p.maximum_size):

                # Slice the image
                sliced_img = cv_slice_img(img, img_resized, cnt, p.perspective_fix)

                # Resize the slice
                if p.scale_factor:
                    sliced_img = cv_resize(sliced_img, scale=p.scale_factor)

                if p.scale_width:
                    sliced_img = cv_resize(sliced_img, w=p.scale_width)

                if p.scale_height:
                    sliced_img = cv_resize(sliced_img, h=p.scale_height)

                # Auto-rotate the slice
                if p.auto_rotate in ["cw", "ccw"]:
                    sliced_img = cv_auto_rotate(sliced_img, p.auto_rotate)

                # Apply filters to slice
                sliced_img = pil_filter_image(sliced_img, filters)

                # Define temporary filename
                filename = "tmp_file_" + random_string(self.name, str(self.size), str(self.slice_count)) + savefile_suffix

                # Save the slice (make sure it does not exist)
                if not os.path.isfile(os.path.join(save_path, filename)):
                    sliced_img.save(os.path.join(save_path, filename), **file_params)
                else:
                    logger.error(f"File already exists: {filename}")
                    return 0

                # Up the counter
                self.slice_count += 1

        # Output warning if no images found
        if not self.slice_count:
            logger.warning(f"[ID:{self.id}] - ({self.name}) - No images found, skipping it..")
            return 0

        return self.slice_count

    # Create test image for GUI
    def create_test_image(self, p):
        # Load image
        img = pil_to_cv(pil_open_image(self.filepath))
        img_resized = cv_resize(img, w=900)

        # Define detection colors (BGR format)
        color_1 = (230, 97, 0)
        color_2 = (93, 58, 155)

        # Detect and draw slices
        for cnt in cv_detect_slices(cv_apply_wt(img_resized, p.white_threshold)):

            # Valid slice detected
            if cv_is_cnt_in_range(img_resized, cnt, p.minimum_size, p.maximum_size):
                self.slice_count += 1
                cv_draw_cnt(img_resized, cnt, color_1, 4)

            # Non-valid slice detected
            else:
                self.false_slice_count += 1
                cv_draw_cnt(img_resized, cnt, color_2, 4)

        # Return detected slices
        img_resized = cv_resize(img_resized, w=min(img_resized.shape[1], p.view_width))
        img_resized = cv_resize(img_resized, h=min(img_resized.shape[0], p.view_height))

        return cv_to_pil(img_resized)

    # Create array for preview images for the GUI
    def create_preview_images(self, p):
        # Array for slices
        preview_images = []

        # Load image
        img = pil_to_cv(pil_open_image(self.filepath))
        img_resized = cv_resize(img, w=900)

        for cnt in cv_detect_slices(cv_apply_wt(img_resized, p.white_threshold)):

            # Make sure the area is larger than minimum area size
            if cv_is_cnt_in_range(img_resized, cnt, p.minimum_size, p.maximum_size):

                # Slice the image
                sliced_img = cv_slice_img(img, img_resized, cnt, p.perspective_fix)

                # Auto-rotate the slice
                if p.auto_rotate in ["cw", "ccw"]:
                    sliced_img = cv_auto_rotate(sliced_img, p.auto_rotate)

                # Append slice
                preview_images.append(sliced_img)

        # Return array of slices
        return preview_images